# -*- coding: utf-8 -*-
# pylint: disable=all
"""genomehubs documentation config."""

import os
import traceback

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
]
source_suffix = ".rst"
master_doc = "index"
project = "genomehubs"
year = "2020"
author = "genomehubs"
author_email = "genomehubs@genomehubs.org"
copyright = "{0}, {1}".format(year, author)
try:
    from pkg_resources import get_distribution

    version = release = get_distribution("genomehubs").version
except Exception:
    traceback.print_exc()
    version = release = "2.2.0"

pygments_style = "trac"
templates_path = ["."]
extlinks = {
    "issue": ("https://github.com/genomehubs/genomehubs/issues/%s", "#"),
    "pr": ("https://github.com/genomehubs/genomehubs/pull/%s", "PR #"),
}
# on_rtd is whether we are on readthedocs.org
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:  # only set the theme if we're building docs locally
    html_theme = "sphinx_rtd_theme"

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
    "**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False



#!/usr/bin/env python3

"""Load configuration from file(s)."""

import os

from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)


def set_common_values(group_options, common_options):
    """Set common options across all option groups."""
    for key, value in common_options.items():
        if key not in group_options:
            group_options[key] = value
        elif isinstance(group_options[key], dict):
            for sub_key, sub_value in common_options[key].items():
                if sub_key not in group_options[key]:
                    group_options[key][sub_key] = sub_value


def update_subsets(flattened_group_options, group_options):
    """Update nested config options."""
    for key, value in group_options.items():
        if isinstance(group_options[key], dict):
            for sub_key, sub_value in group_options[key].items():
                if isinstance(group_options[key][sub_key], dict):
                    for sub_sub_key, sub_sub_value in group_options[key][
                        sub_key
                    ].items():
                        if isinstance(sub_sub_value, str) and sub_sub_value.startswith(
                            "~"
                        ):
                            sub_sub_value = os.path.expanduser(sub_sub_value)
                        elif (
                            not isinstance(sub_sub_value, list)
                            and sub_sub_key == "root"
                        ):
                            sub_sub_value = [sub_sub_value]
                        flattened_group_options[
                            "%s-%s-%s" % (key, sub_key, sub_sub_key)
                        ] = sub_sub_value
                else:
                    if isinstance(sub_value, str) and sub_value.startswith("~"):
                        sub_value = os.path.expanduser(sub_value)
                    flattened_group_options["%s-%s" % (key, sub_key)] = sub_value
        else:
            if isinstance(value, str) and value.startswith("~"):
                value = os.path.expanduser(value)
            flattened_group_options[key] = value


def load_config(options, file):
    """Load configuration from file."""
    new_options = tofile.load_yaml(file)
    if "common" in new_options:
        for group_name, group_options in new_options.items():
            if group_name != "common":
                set_common_values(group_options, new_options["common"])
    for group_name, group_options in new_options.items():
        if group_name not in options:
            options[group_name] = {}
        nested_group_options = options[group_name]
        update_subsets(nested_group_options, group_options)
    return options


def enforce_lists(options):
    """Ensure specified options are lists."""
    keys = {"es-host"}
    for key in keys:
        if key in options:
            if not isinstance(options[key], list):
                options[key] = [options[key]]


def config(group, **kwargs):
    """Load configuration."""
    LOGGER.info("Loading configuration options")
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    dist_file = os.path.join(script_dir, "config", "dist.config.yaml")
    user_file = os.path.join(script_dir, "config", "config.yaml")
    options = {}
    if os.path.exists(dist_file):
        options = load_config(options, dist_file)
    else:
        print(dist_file)
    if os.path.exists(user_file):
        options = load_config(options, user_file)
    if "--config-file" in kwargs:
        for file in kwargs["--config-file"]:
            if os.path.exists(file):
                options = load_config(options, file)
    if group not in options:
        options[group] = {}
    for k, v in kwargs.items():
        if v:
            k = k.lstrip("--")
            if isinstance(v, str):
                if v.startswith("~"):
                    v = os.path.expanduser(v)
            if k in options[group]:
                options[group][k] = v
            elif not k == group:
                options[group][k] = v
    enforce_lists(options)
    return options



#!/usr/bin/env python3
"""Hub functions."""

import csv
import os
import re
import sys
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

from tolkein import tofile
from tolkein import tolog

LOGGER = tolog.logger(__name__)
MIN_INTEGER = -(2 ** 31)
MAX_INTEGER = 2 ** 31 - 1


def setup(opts):
    """Set up directory for this GenomeHubs instance and handle reset."""
    Path(opts["hub-path"]).mkdir(parents=True, exist_ok=True)
    return True


def load_types(name, *, part="types"):
    """Read a types file from the templates directory."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    types = None
    try:
        types_file = os.path.join(script_dir, "templates", "%s.%s.yaml" % (name, part))
        types = tofile.load_yaml(types_file)
        try:
            for key, value in types.items():
                try:
                    enum = set(
                        [str(option).lower() for option in value["constraint"]["enum"]]
                    )
                    value["constraint"]["enum"] = enum
                except KeyError:
                    pass
        except AttributeError:
            pass
    except Exception:
        pass
    if types and "file" in types and "attributes" in types:
        for key, value in types["file"].items():
            for attr in types["attributes"].values():
                if key not in ("format", "header", "name") and key not in attr:
                    attr[key] = value
    return types


def index_templator(parts, opts):
    """Index template helper function."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    mapping_file = os.path.join(script_dir, "templates", parts[0] + ".json")
    types = load_types(parts[0])
    names = load_types(parts[0], part="names")
    template = {
        "name": parts[0],
        "index_name": opts["hub-separator"].join(parts),
        "mapping": tofile.load_yaml(mapping_file),
        "types": types,
        "names": names,
    }
    return template


def add_names_to_types(names, types):
    """Add names field meta to type field meta."""
    sources = 0
    if types is not None:
        types = deepcopy(types)
        sources += 1
    elif names is not None:
        types = deepcopy(names)
        sources += 1
    if sources == 2:
        for group, entries in names.items():
            if group not in types:
                types[group] = deepcopy(entries)
            else:
                for field, attrs in entries.items():
                    if isinstance(attrs, dict):
                        if field not in types[group]:
                            types[group][field] = deepcopy(attrs)
                        elif types[group][field]["header"] != attrs["header"]:
                            types[group]["names_%s" % field] = deepcopy(attrs)
    return types


def order_parsed_fields(parsed, types, names=None):
    """Order parsed fields using a template file."""
    columns = {}
    fields = {}
    ctr = 0
    types = add_names_to_types(names, types)
    for group, entries in types.items():
        for field, attrs in entries.items():
            header = False
            try:
                for key, value in attrs.items():
                    if key == "index":
                        if value not in columns:
                            columns.update({value: field})
                            fields.update({field: value})
                    elif key == "header":
                        header = value
                if header and header not in fields:
                    columns.update({ctr: header})
                    fields.update({header: ctr})
                    ctr += 1
            except AttributeError:
                pass
    order = [x[0] for x in sorted(fields.items(), key=lambda x: x[1])]
    data = [order]
    for entry in parsed:
        row = [entry.get(field, "None") for field in order]
        data.append(row)
    return data


def post_search_scripts(es):
    """POST ElasticSearch search scripts."""
    script_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    search_script_dir = os.path.join(script_dir, "templates", "scripts")
    directory = os.fsencode(search_script_dir)
    for script_file in os.listdir(directory):
        filename = os.fsdecode(script_file)
        if filename.endswith(".json"):
            body = tofile.load_yaml(os.path.join(search_script_dir, filename))
            script_id = filename.replace(".json", "")
            with tolog.DisableLogger():
                es.put_script(id=script_id, body=body)


def byte_type_constraint(value):
    """Test byte_type constraint."""
    if value >= -128 and value <= 128:
        return True
    return False


def integer_type_constraint(value):
    """Test integer_type constraint."""
    if value >= MIN_INTEGER and value <= MAX_INTEGER:
        return True
    return False


def short_type_constraint(value):
    """Test short_type constraint."""
    if value >= -32768 and value <= 32767:
        return True
    return False


def min_value_constraint(value, limit):
    """Test minimum value constraint."""
    if value >= limit:
        return True
    return False


def max_value_constraint(value, limit):
    """Test maximum value constraint."""
    if value <= limit:
        return True
    return False


def enum_constraint(value, enum):
    """Test value in predefined set constraint."""
    if str(value).lower() in enum:
        return True
    return False


def test_constraint(value, constraint):
    """Test value against constraint."""
    constraints = {
        "byte": byte_type_constraint,
        "integer": integer_type_constraint,
        "short": short_type_constraint,
        "min": min_value_constraint,
        "max": max_value_constraint,
        "enum": enum_constraint,
    }
    for key in constraint:
        if key in constraints:
            if not constraints[key](value, constraint[key]):
                return False
    return True


def convert_lat_lon(location):
    """Convert lat and lon to array notation."""
    if isinstance(location, list):
        return ",".join(location)
    if location.endswith(("E", "W")):
        sign = {"N": "", "E": "", "S": "-", "W": "-"}
        parts = re.split(r"\s*([NESW])\s*", location)
        string = "%s%s,%s%s" % (sign[parts[1]], parts[0], sign[parts[3]], parts[2])
        return string
    if location:
        return None
    return ""


def convert_to_type(key, value, to_type):
    """Convert values to type."""
    if to_type in {"byte", "integer", "long", "short"}:
        try:
            value = int(value)
        except ValueError:
            value = None
    elif to_type in {
        "double",
        "float",
        "half_float",
        "1dp",
        "2dp",
        "3dp",
        "4dp",
    }:
        try:
            value = float(value)
        except ValueError:
            value = None
    elif to_type == "geo_point":
        value = convert_lat_lon(value)
    else:
        value = str(value)
    # if value is None:
    #     LOGGER.warning(
    #         "%s value %s is not a valid %s", key, str(value), to_type,
    #     )
    return value


def calculator(value, operation):
    """Template-based calculator."""
    value1, operator, value2 = operation.split(" ")
    if value1 == "{}":
        value1 = float(value)
        value2 = float(value2)
    elif value2 == "{}":
        value2 = float(value)
        value1 = float(value1)
    if operator == "+":
        return value1 + value2
    elif operator == "-":
        return value1 - value2
    elif operator == "*":
        return value1 * value2
    elif operator == "/":
        try:
            return value1 / value2
        except ZeroDivisionError:
            return None
    elif operator == "%":
        if operator == "%":
            return value1 % value2
    else:
        return None


def validate_values(values, key, types):
    """Validate values."""
    validated = []
    key_type = types[key]["type"]
    for value in values:
        if "function" in types[key]:
            try:
                value = calculator(value, types[key]["function"])
            except ValueError:
                continue
        value = convert_to_type(key, value, key_type)
        if value is None:
            continue
        if isinstance(value, str):
            if not value:
                continue
            try:
                value = types[key]["translate"][value]
            except KeyError:
                pass
        try:
            valid = test_constraint(value, key_type)
            if valid:
                valid = test_constraint(value, types[key]["constraint"])
        except KeyError:
            valid = True
        if valid:
            validated.append(value)
        elif value:
            LOGGER.warning("%s is not a valid %s value", str(value), key)
    return validated


def apply_value_template(prop, value, attribute, *, taxon_types, has_taxon_data):
    """Set value using template."""
    template = re.compile(r"^(.*?\{\{)(.+)(\}\}.*)$")
    new_prop = prop.replace("taxon_", "")
    match = template.match(str(value))
    if match:
        groups = match.groups()
        if groups[1] and groups[1] in attribute:
            has_taxon_data = True
            new_value = attribute[groups[1]]
            if groups[0] != "{{":
                new_value = "%s%s" % (
                    groups[0].replace("{{", ""),
                    str(new_value),
                )
            if groups[2] != "}}":
                new_value = "%s%s" % (
                    str(new_value),
                    groups[2].replace("}}", ""),
                )
            taxon_types.update({new_prop: new_value})
            taxon_types.update({"group": "taxon"})
            if prop in taxon_types:
                del taxon_types[prop]
    else:
        has_taxon_data = True
        taxon_types.update({new_prop: value})
        taxon_types.update({"group": "taxon"})
        if prop in taxon_types:
            del taxon_types[prop]
    return has_taxon_data


def add_attributes(
    entry, types, *, attributes=None, source=None, attr_type="attributes", meta=None
):
    """Add attributes to a document."""
    if attributes is None:
        attributes = []
    taxon_attributes = []
    taxon_types = {}
    if meta is None:
        meta = {}
    attribute_values = {}
    for key, values in entry.items():
        if key in types:
            if not isinstance(values, list):
                values = [values]
            validated = validate_values(values, key, types)
            # TODO: handle invalid values
            if validated:
                if len(validated) == 1:
                    validated = validated[0]
                if attr_type == "attributes":
                    attribute = {"key": key, "%s_value" % types[key]["type"]: validated}
                    attribute_values.update({key: attribute})
                else:
                    attribute = {"identifier": validated, "class": key}
                attribute.update(meta)
                if "source" in types[key]:
                    attribute.update({"source": types[key]["source"]})
                elif source is not None:
                    attribute.update({"source": source})
                attributes.append(attribute)
    if attribute_values:
        for attribute in attributes:
            has_taxon_data = False
            taxon_attribute = {**attribute, "source_index": "assembly"}
            taxon_attribute_types = {**types[attribute["key"]]}
            # taxon_props = []
            for prop, value in types[attribute["key"]].items():
                if prop.startswith("taxon_"):
                    has_taxon_data = apply_value_template(
                        prop,
                        value,
                        attribute,
                        taxon_types=taxon_attribute_types,
                        has_taxon_data=has_taxon_data,
                    )
                    # taxon_props.append(prop)
            # for prop in taxon_props:
            #     del types[attribute["key"]][prop]
            if has_taxon_data:
                taxon_attribute.update({"key": taxon_attribute_types["name"]})
                # taxon_attribute.update({"name": taxon_attribute_types["key"]})
                taxon_attributes.append(taxon_attribute)
                taxon_types.update(
                    {taxon_attribute_types["name"]: taxon_attribute_types}
                )
    return attributes, taxon_attributes, taxon_types


def chunks(arr, n):
    """Yield successive n-sized chunks from arr."""
    for i in range(0, len(arr), n):
        yield arr[i : i + n]


def add_attribute_values(existing, new, *, raw=True):
    """Add attribute values to records."""
    indices = {}
    for index, entry in enumerate(existing):
        indices[entry["key"]] = index
    index = len(existing)
    for group in new:
        if not isinstance(group, list):
            group = [group]
        for entry in group:
            if raw:
                arr = []
                if indices and entry["key"] in indices:
                    try:
                        arr = existing[indices[entry["key"]]]["values"]
                    except KeyError:
                        LOGGER.error(
                            "Unable to import values to an attribute (%s) that has already been filled",
                            entry["key"],
                        )
                        exit(1)
                else:
                    existing.append({"key": entry["key"], "values": arr})
                    indices[entry["key"]] = index
                    index += 1
                del entry["key"]
                arr.append(entry)
            else:
                existing.append(
                    {
                        **entry,
                        "count": 1,
                        "aggregation_method": "unique",
                        "aggregation_source": "direct",
                    }
                )


def strip_comments(data, types):
    """Strip comment lines from a file stream."""
    comment_chars = {"#"}
    if "file" in types and "comment" in types["file"]:
        comment_chars.update(set(types["file"]["comment"]))
    for row in data:
        if row[0] in comment_chars:
            continue
        yield row


def process_names_file(types, names_file):
    """Process a taxon names file."""
    data = tofile.open_file_handle(names_file)
    names = defaultdict(dict)
    if data is None:
        return names
    delimiters = {"csv": ",", "tsv": "\t"}
    rows = csv.reader(
        strip_comments(data, types),
        delimiter=delimiters[types["file"]["format"]],
        quotechar='"',
    )
    next(rows)
    for row in rows:
        name = row[3] if len(row) > 3 else row[1]
        names[row[2]][row[1]] = {"name": name, "taxon_id": row[0]}
    return names


def validate_types_file(types_file, dir_path):
    """Validate types file."""
    LOGGER.info("Validating YAML file %s", types_file)
    try:
        types = tofile.load_yaml(str(types_file.resolve()))
    except Exception:
        LOGGER.error("Unable to open types file %s", str(types_file.resolve()))
        sys.exit(1)
    if "taxonomy" not in types:
        LOGGER.error("Types file contains no taxonomy information")
        sys.exit(1)
    if "file" not in types or "name" not in types["file"]:
        LOGGER.error("No data file name in types file")
        sys.exit(1)
    if "attributes" in types:
        for entry in types["attributes"].values():
            enum = entry.get("constraint", {}).get("enum", [])
            if enum:
                entry["constraint"]["enum"] = [value.lower() for value in enum]
    defaults = {"attributes": {}, "metadata": {}}
    for key, value in types["file"].items():
        if key.startswith("display") or key.startswith("taxon"):
            defaults["attributes"].update({key: value})
        elif key.startswith("source"):
            defaults["attributes"].update({key: value})
            defaults["metadata"].update({key: value})
    types.update({"defaults": defaults})
    datafile = Path(dir_path) / types["file"]["name"]
    data = tofile.open_file_handle(datafile)
    if data is None:
        LOGGER.error("Data file '%s' could not de opened for reading", datafile)
        sys.exit(1)
    names = process_names_file(types, Path(dir_path) / "names" / types["file"]["name"])
    return types, data, names


def set_xrefs(taxon_names, types, row, *, meta=None):
    """Set xrefs for taxon_names."""
    if meta is None:
        meta = {}
    names = []
    for name_class, value in taxon_names.items():
        taxon = {"name": value, "class": name_class}
        if "xref" in types[name_class] and types[name_class]["xref"]:
            if "source" in meta:
                taxon.update({"source": meta["source"]})
            if "source_stub" in meta:
                taxon.update({"source_stub": meta["source_stub"]})
        names.append(taxon)
    return names


def set_row_defaults(types, data):
    """Set default values for a row."""
    for key in types["defaults"].keys():
        if key in types:
            for entry in types[key].values():
                entry = {
                    **types["defaults"][key],
                    **entry,
                }
        elif key == "metadata":
            data["metadata"] = {**types["defaults"]["metadata"]}


def process_row_values(row, types, data):
    """Process row values."""
    for group in data.keys():
        if group in types:
            for key, meta in types[group].items():
                if "index" not in meta:
                    continue
                try:
                    if isinstance(meta["index"], list):
                        char = meta.get("join", "")
                        values = [row[i] for i in meta["index"]]
                        if all(values):
                            value = char.join(values)
                        else:
                            continue
                    else:
                        value = row[meta["index"]]
                    if "separator" in meta and any(
                        sep in value for sep in meta["separator"]
                    ):
                        separator = "|".join(meta["separator"])
                        data[group][key] = re.split(rf"\s*{separator}\s*", value)
                    elif value is not None and value != "None":
                        data[group][key] = value
                except Exception as err:
                    LOGGER.warning("Cannot parse row '%s'" % str(row))
                    raise err


def process_row(types, names, row):
    """Process a row of data."""
    data = {
        "attributes": {},
        "identifiers": {},
        "metadata": {},
        "taxon_names": {},
        "taxonomy": {},
        "taxon_attributes": {},
    }
    set_row_defaults(types, data)
    process_row_values(row, types, data)
    taxon_data = {}
    taxon_types = {}
    if "is_primary_value" in data["metadata"]:
        try:
            data["metadata"]["is_primary_value"] = bool(
                int(data["metadata"]["is_primary_value"])
            )
        except ValueError:
            data["metadata"]["is_primary_value"] = False
    assembly_id = None
    if (
        "identifiers" in data
        and data["identifiers"]
        and "assembly_id" in data["identifiers"]
        and data["identifiers"]["assembly_id"]
    ):
        assembly_id = data["identifiers"]["assembly_id"]
    for attr_type in list(["attributes", "identifiers"]):
        if attr_type in data and data[attr_type]:
            (
                data[attr_type],
                taxon_data[attr_type],
                taxon_types[attr_type],
            ) = add_attributes(
                data[attr_type],
                types[attr_type],
                attr_type=attr_type,
                meta=data["metadata"],
            )
        else:
            data[attr_type] = []
    if assembly_id is not None:
        if "attributes" in taxon_data and taxon_data["attributes"]:
            for attr in taxon_data["attributes"]:
                attr.update(
                    {
                        "source_index": "assembly",
                        "source_id": assembly_id,
                    }
                )
    if data["taxon_names"]:
        data["taxon_names"] = set_xrefs(
            data["taxon_names"], types["taxon_names"], row, meta=data["metadata"]
        )
    if data["taxonomy"] and names:
        for key in names.keys():
            if key in data["taxonomy"]:
                if data["taxonomy"][key] in names[key]:
                    data["taxonomy"]["_taxon_id"] = names[key][data["taxonomy"][key]][
                        "taxon_id"
                    ]
                    data["taxonomy"][key] = names[key][data["taxonomy"][key]]["name"]
    return data, taxon_data, taxon_types.get("attributes", {})


def set_column_indices(types, header):
    """Use header to set indices for named columns."""
    headers = {title: index for index, title in enumerate(header)}
    for section, entries in types.items():
        for key, value in entries.items():
            if isinstance(value, dict):
                if "header" in value:
                    index = headers.get(value["header"], None)
                    if index is not None:
                        value.update({"index": index})


def write_imported_rows(rows, opts, *, types, header=None, label="imported"):
    """Write imported rows to processed file."""
    file_key = "%s-exception" % opts["index"]
    dir_key = "%s-dir" % opts["index"]
    if file_key in opts and opts[file_key]:
        outdir = opts[file_key]
    else:
        outdir = "%s/%s" % (opts[dir_key], label)
    os.makedirs(outdir, exist_ok=True)
    outfile = "%s/%s" % (outdir, types["file"]["name"])
    data = []
    header_len = 0
    if header is not None:
        data.append(header)
        header_len = 1
    if isinstance(rows, dict):
        for row_set in rows.values():
            for row in row_set:
                data.append(row)
    else:
        for row in rows:
            data.append(row)
    LOGGER.info(
        "Writing %d records to %s file '%s'", len(data) - header_len, label, outfile
    )
    tofile.write_file(outfile, data)


def write_spellchecked_taxa(spellings, opts, *, types):
    """Write spellchecked taxa to file."""
    dir_key = "%s-dir" % opts["index"]
    filepath = Path(types["file"]["name"])
    extensions = "".join(filepath.suffixes)
    file_basename = str(filepath).replace(extensions, "")
    dirs = {
        "spellcheck": "exceptions",
        "synonym": "imported",
    }
    for group in dirs.keys():
        taxa = []
        for name, obj in spellings[group].items():
            taxa.append([obj["taxon_id"], name, obj["rank"]] + obj["matches"])
        if taxa:
            outdir = "%s/%s" % (opts[dir_key], dirs[group])
            os.makedirs(outdir, exist_ok=True)
            outfile = "%s/%s" % (outdir, "%s.spellcheck.tsv" % file_basename)
            LOGGER.info(
                "Writing %d %s suggestions to spellcheck file '%s'",
                len(taxa),
                group,
                outfile,
            )
            tofile.write_file(
                outfile, [["taxon_id", "input", "rank", "suggested"]] + taxa
            )


def write_imported_taxa(taxa, opts, *, types):
    """Write imported taxa to file."""
    imported = []
    file_key = "%s-exception" % opts["index"]
    dir_key = "%s-dir" % opts["index"]
    filepath = Path(types["file"]["name"])
    extensions = "".join(filepath.suffixes)
    file_basename = str(filepath).replace(extensions, "")
    for name, arr in taxa.items():
        prefix = "#" if len(arr) > 1 else ""
        for obj in arr:
            imported.append(
                ["%s%s" % (prefix, str(obj["taxon_id"])), name, obj["rank"]]
            )
    if imported:
        if file_key in opts and opts[file_key]:
            outdir = opts[file_key]
        else:
            outdir = "%s/imported" % opts[dir_key]
        os.makedirs(outdir, exist_ok=True)
        outfile = "%s/%s" % (outdir, "%s.taxon_ids.tsv" % file_basename)
        LOGGER.info(
            "Writing %d taxon_ids to imported file '%s'",
            len(imported),
            outfile,
        )
        tofile.write_file(outfile, [["taxon_id", "input", "rank"]] + imported)



#!/usr/bin/env python3

"""Taxonomy methods."""

from pathlib import Path

from tolkein import tofetch
from tolkein import tolog
from tolkein import totax

from .hub import index_templator

LOGGER = tolog.logger(__name__)


def index_template(name, opts):
    """Index template (includes name, mapping and types)."""
    parts = ["taxonomy", name, opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def files_exist(expected_files, path):
    """Test if expected files already exist."""
    for filename in expected_files:
        if not (path / filename).exists():
            return False
    return True


def confirm_index_opts(taxonomy_name, opts):
    """Confirm expected keys are present in opts for indexing."""
    file_key = "taxonomy-%s-file" % taxonomy_name
    url_key = "taxonomy-%s-url" % taxonomy_name
    for key in {"taxonomy-path", file_key}:
        if key not in opts:
            LOGGER.warning("Unable to index %s, '%s' not specified", taxonomy_name, key)
            return False
    taxonomy_path = Path("%s/%s" % (opts["taxonomy-path"], taxonomy_name))
    taxonomy_path.mkdir(parents=True, exist_ok=True)
    if url_key not in opts:
        if not files_exist(opts[file_key], taxonomy_path):
            LOGGER.warning(
                "Unable to index %s, '%s' not specified and files not found at '%s'",
                taxonomy_name,
                url_key,
                str(taxonomy_path),
            )
            return False
    return True


def index(taxonomy_name, opts):
    """Index a taxonomy."""
    if "taxonomy-%s-tree" % taxonomy_name in opts:
        LOGGER.warning(
            "Unable to import %s. Trees are not yet supported as a taxonomy type",
            taxonomy_name,
        )
        return
    if not confirm_index_opts(taxonomy_name, opts):
        return
    LOGGER.info("Indexing %s", taxonomy_name)
    template = index_template(taxonomy_name, opts)
    taxonomy_path = Path("%s/%s" % (opts["taxonomy-path"], taxonomy_name))
    file_key = "taxonomy-%s-file" % taxonomy_name
    if not files_exist(opts[file_key], taxonomy_path):
        LOGGER.info(
            "Fetching %s taxdump and extracting to %s",
            taxonomy_name,
            str(taxonomy_path),
        )
        tofetch.fetch_tar(
            url=opts["taxonomy-%s-url" % taxonomy_name], path=str(taxonomy_path)
        )
        for file in opts[file_key]:
            for p in taxonomy_path.rglob(file):
                p.rename(taxonomy_path / p.name)
    else:
        LOGGER.info(
            "Using existing %s taxdump at %s", taxonomy_name, str(taxonomy_path)
        )
    root_key = "taxonomy-%s-root" % taxonomy_name
    root = opts.get(root_key, None)
    stream = totax.parse_taxonomy(taxonomy_name, str(taxonomy_path), root)
    return template, stream



#!/usr/bin/env python3
"""Elasticsearch functions."""

import logging
import os
import platform
import signal
import sys
import time
from pathlib import Path
from subprocess import PIPE
from subprocess import Popen

import ujson
from elasticsearch import Elasticsearch
from elasticsearch import NotFoundError
from elasticsearch import client
from elasticsearch import helpers
from tolkein import tofile
from tolkein import tolog
from tqdm import tqdm

LOGGER = tolog.logger(__name__)


def test_connection(opts, *, log=False):
    """Test connection to Elasticsearch."""
    connected = False
    hosts = opts["es-host"]
    with tolog.DisableLogger():
        try:
            es = Elasticsearch(
                hosts=hosts, timeout=1800, max_retries=10, retry_on_timeout=True
            )
            connected = es.info()
        except Exception:
            pass
    if not connected:
        message = "Could not connect to Elasticsearch at '%s'" % ", ".join(hosts)
        if log:
            LOGGER.error(message)
            sys.exit(1)
        return False
    if log:
        LOGGER.info("Connected to Elasticsearch")
    return es


def write_elastic_yaml(path, opts):
    """Write Elasticsearch YAML config file."""
    try:
        port = int(opts["es-host"][0].split(":")[1])
    except ValueError:
        port = 9200
    Path("%s/data" % opts["es-path"]).mkdir(parents=True, exist_ok=True)
    Path("%s/logs" % opts["es-path"]).mkdir(parents=True, exist_ok=True)
    data = {
        "path.data": "%s/data" % opts["es-path"],
        "path.logs": "%s/logs" % opts["es-path"],
        "http.port": port,
    }
    tofile.write_file(path, data)
    return port


def start_es_docker(opts):
    """Use Elasticsearch Docker image."""
    LOGGER.error("Elasticsearch Docker options have not been implemented yet")
    return None


def start_es_binary(opts):
    """Use ElasticSearch binary."""
    es_full_version = "%s-%s-%s" % (
        opts["es-version"],
        platform.system().lower(),
        platform.machine(),
    )
    es_tarfile = "%s.tar.gz" % es_full_version
    es_bin_path = Path(opts["es-path"]) / opts["es-version"].replace("-oss", "") / "bin"
    if not es_bin_path.exists():
        es_url = "%s/%s" % (opts["es-url"], es_tarfile)
        LOGGER.info("Fetching ElasticSearch from '%s'", es_url)
    LOGGER.info("Writing ElasticSearch config")
    port = write_elastic_yaml("%s/config/elasticsearch.yml" % opts["es-path"], opts)
    LOGGER.info("Starting ElasticSearch on port %d", port)
    process = Popen(
        "%s/elasticsearch" % es_bin_path, stdout=PIPE, stderr=PIPE, encoding="ascii"
    )
    LOGGER.info("Starting ElasticSearch with pid %d", process.pid)
    es = test_connection(opts)
    pbar = tqdm(total=int(opts["es-timeout"]))
    for _i in range(int(opts["es-timeout"])):
        pbar.update(1)
        time.sleep(1)
        es = test_connection(opts)
        if es:
            break
    pbar.close()
    if es:
        LOGGER.info("ElasticSearch running on port %d", port)
        return es
    LOGGER.error(
        "Unable to start ElasticSearch in %d seconds",
        int(opts["es-timeout"]),
    )
    LOGGER.info("Consider increasing es-timeout")
    LOGGER.info("Stopping ElasticSearch process")
    try:
        os.kill(process.pid, signal.SIGTERM)
    except ProcessLookupError:
        pass
    sys.exit(1)


def launch_es(opts, log=True):
    """Launch ElasticSearch."""
    es = test_connection(opts)
    if es:
        if log:
            LOGGER.info("ElasticSearch is already running")
    if not es:
        if any(host.startswith(("localhost", "127.0.0.1")) for host in opts["es-host"]):
            # Start Elasticsearch
            if "docker-contain" in opts and "es" in opts["docker-contain"]:
                es = start_es_docker(opts)
            else:
                es = start_es_binary(opts)
        else:
            LOGGER.error("Unable to start ElasticSearch on remote host")
            LOGGER.info(
                "Specify localhost or ensure ElasticSearch is running remotely before executing this command"
            )
            sys.exit(1)
        if not es:
            LOGGER.error("Unable to start ElasticSearch")
            sys.exit(1)
    return es


def index_exists(es, index_name):
    """Test if Elasticsearch index exists."""
    es_client = client.IndicesClient(es)
    with tolog.DisableLogger():
        res = es_client.exists(index_name)
    return res


def index_create(es, index_name):
    """Create an Elasticsearch index if it does not already exist."""
    es_client = client.IndicesClient(es)
    res = index_exists(es, index_name)
    if not res:
        with tolog.DisableLogger():
            res = es_client.create(index_name)
    return res


def load_mapping(es, mapping_name, mapping):
    """Load index mapping template into Elasticsearch."""
    es_client = client.IndicesClient(es)
    with tolog.DisableLogger():
        res = es_client.put_template(name=mapping_name, body=mapping)
    return res


def index_stream(es, index_name, stream, *, _op_type="index", log=True, dry_run=False):
    """Load bulk entries from stream into Elasticsearch index."""
    # LOGGER.info("Indexing bulk entries to %s", index_name)
    if _op_type == "index":
        actions = (
            {
                "_index": index_name,
                "_id": entry_id,
                "_source": entry,
                "_op_type": _op_type,
            }
            for entry_id, entry in stream
        )
    elif _op_type == "update":
        actions = (
            {"_index": index_name, "_id": entry_id, "doc": entry, "_op_type": _op_type}
            for entry_id, entry in stream
        )

    def dry_run_iterator(es, actions):
        """Alternate iterator for dry run."""
        for action in actions:
            yield True, {}

    try:
        tracer = logging.getLogger("elasticsearch")
        tracer.setLevel(logging.ERROR)
        if dry_run:
            iterator = dry_run_iterator(es, actions)
        else:
            iterator = helpers.streaming_bulk(es, actions)
        success = 0
        failed = 0
        if log:
            iterator = tqdm(iterator, unit=" records", unit_scale=True)
        for ok, response in iterator:
            if ok:
                success += 1
            else:
                failed += 1
    except Exception as err:
        raise err

    return success, failed


def stream_template_search_results(es, *, index, body, size=10):
    """Stream results of a template search."""
    body["params"].update({"size": size})
    with tolog.DisableLogger():
        res = es.search_template(
            index=index, body=body, rest_total_hits_as_int=True, scroll="90m"
        )
    scroll_id = res["_scroll_id"]
    count = res["hits"]["total"]
    for hit in res["hits"]["hits"]:
        yield hit
    offset = size
    while offset < count:
        with tolog.DisableLogger():
            res = es.scroll(
                rest_total_hits_as_int=True, scroll="90m", scroll_id=scroll_id
            )
        for hit in res["hits"]["hits"]:
            yield hit
        offset += size
    with tolog.DisableLogger():
        es.clear_scroll(scroll_id=scroll_id)


def query_keyword_value_template(es, template_name, keyword, values, index, opts=None):
    """Run query using a by_keyword_value template."""
    if not index_exists(es, index):
        return None
    if opts is None:
        opts = {"keyword": "keyword", "value": "value"}
    multisearch = False
    body = ""
    if isinstance(values, list):
        multisearch = True
    else:
        values = [values]
    for value in values:
        if multisearch:
            body += "{}\n"
        body += ujson.dumps(
            {
                "id": template_name,
                "params": {opts["keyword"]: keyword, opts["value"]: value},
            }
        )
        body += "\n"
    res = None
    if multisearch:
        with tolog.DisableLogger():
            res = es.msearch_template(body=body, index=index)
    else:
        with tolog.DisableLogger():
            res = es.search_template(body=body, index=index)
    return res


def query_value_template(es, template_name, values, index):
    """Run query using a by_value template."""
    if not index_exists(es, index):
        return None
    multisearch = False
    body = ""
    if isinstance(values, list):
        multisearch = True
    else:
        values = [values]
    if not values:
        return None
    for value in values:
        if multisearch:
            body += "{}\n"
        body += ujson.dumps({"id": template_name, "params": {"value": value}})
        body += "\n"
    res = None
    if multisearch:
        with tolog.DisableLogger():
            res = es.msearch_template(body=body, index=index)
    else:
        with tolog.DisableLogger():
            res = es.search_template(body=body, index=index)
    return res


def document_by_id(es, ids, index):
    """Get indexed documents by ID."""
    if not index_exists(es, index):
        return None
    multisearch = False
    if not ids:
        return None
    if isinstance(ids, list):
        multisearch = True
    res = None
    try:
        if multisearch:
            with tolog.DisableLogger():
                res = es.mget(body={"ids": ids}, index=index)
                ret = {}
                for result in res["docs"]:
                    if "found" in result and result["found"]:
                        ret.update({result["_id"]: result["_source"]})
                res = ret
        else:
            with tolog.DisableLogger():
                res = es.get(id=ids, index=index)
                if "found" in res and res["found"]:
                    res = {res["_id"]: res["_source"]}
                else:
                    res = None
    except NotFoundError:
        res = None
    return res


class EsQueryBuilder:
    """Class for building ElasticSearch queries."""

    def __init__(self):
        """Init EsQueryBuilder class."""
        self._parts = []
        self._includes = []
        self._excludes = []
        self._query = {}
        return None

    def es_and(self):
        """AND query."""
        self.es_bool()
        return self

    def es_or(self):
        """OR query."""
        self.es_bool("should")
        return self

    def es_bool(self, bool_type="filter"):
        """Bool query."""
        obj = {"bool": {bool_type: self._parts[:]}}
        self._parts = [obj]
        return self

    def es_nested(self, path, bool_type="filter"):
        """Nested query."""
        self.es_bool(bool_type)
        obj = {"nested": {"path": path, "query": self._parts[0]}}
        self._parts = [obj]
        return self

    def es_nested_or(self, path):
        """Nested query."""
        self.es_nested(path, bool_type="filter")
        return self

    def es_nested_and(self, path):
        """Nested query."""
        self.es_nested(path)
        return self

    def es_range(self, key, limits, *, inclusive=True):
        """Range query."""
        obj = {}
        if isinstance(limits, list):
            if limits[0] is not None:
                if inclusive:
                    obj.update({"gte": limits[0]})
                else:
                    obj.update({"gt": limits[0]})
            if len(limits) > 1 and limits[1] is not None:
                if inclusive:
                    obj.update({"lte": limits[1]})
                else:
                    obj.update({"lt": limits[1]})
        else:
            obj.update({"gte": limits[0], "lte": limits[1]})
        self._parts.append({"range": {key: obj}})
        return self

    def es_match(self, key, value):
        """Match query."""
        self._parts.append({"match": {key: str(value)}})
        return self

    def es_include(self, keys):
        """Add fields to include list."""
        if not isinstance(keys, list):
            keys = [keys]
        for key in keys:
            if key not in self._includes:
                self._includes.append(key)
        return self

    def es_exclude(self, keys):
        """Add fields to exclude list."""
        if not isinstance(keys, list):
            keys = [keys]
        for key in keys:
            if key not in self._excludes:
                self._excludes.append(key)
        return self

    def write(self):
        """Return query."""
        if self._parts:
            if len(self._parts) > 1:
                self.es_bool()
            query = {"query": self._parts[0]}
            source = {}
            if self._includes:
                source.update({"includes": self._includes})
            if self._excludes:
                source.update({"excludes": self._excludes})
            if source:
                query.update({"_source": source})
            return query
        return None

    def string(self):
        """Return query as string."""
        return ujson.dumps(self.write())



#!/usr/bin/env python3
"""
GenomeHubs.

Usage: genomehubs [<command>] [<args>...] [-h|--help] [--version]

Commands:
    index           index a file
    init            initialise a new genomehub
    launch          launch a web service
    parse           parse a local or remote data source
    remove          remove one or more indices from a genomehub
    run             run an analysis
    search          search indexed files
    -h, --help      show this
    -v, --version   show version number
See 'genomehubs <command> --help' for more information on a specific command.
"""

import sys

from docopt import DocoptExit
from docopt import docopt
from pkg_resources import working_set
from tolkein import tolog

from .lib.version import __version__

LOGGER = tolog.logger(__name__)


def cli():
    """Entry point."""
    if len(sys.argv) > 1:
        try:
            args = docopt(__doc__, help=False, version=__version__)
        except DocoptExit:
            args = {"<command>": sys.argv[1]}
        if args["<command>"]:
            # load <command> from entry_points
            for entry_point in working_set.iter_entry_points("genomehubs.subcmd"):
                if entry_point.name == args["<command>"]:
                    subcommand = entry_point.load()
                    sys.exit(subcommand())
            LOGGER.error("'genomehubs %s' is not a valid command", args["<command>"])
            sys.exit(1)
    print(__doc__)
    raise DocoptExit



"""Configuration parameters for pytest."""
# pylint: disable=unused-import
import sys
from os.path import abspath
from os.path import dirname as d
from os.path import join

from pytest_mock import mocker



#!/usr/bin/env python3

"""Taxon methods."""

import sys
from collections import defaultdict

from tolkein import tolog
from tqdm import tqdm

from .es_functions import EsQueryBuilder
from .es_functions import document_by_id
from .es_functions import index_stream
from .es_functions import query_keyword_value_template
from .es_functions import query_value_template
from .es_functions import stream_template_search_results
from .hub import add_attribute_values
from .hub import chunks
from .hub import index_templator
from .hub import write_imported_rows
from .taxonomy import index_template as taxonomy_index_template

LOGGER = tolog.logger(__name__)


def index_template(taxonomy_name, opts):
    """Index template (includes name, mapping and types)."""
    parts = ["taxon", taxonomy_name, opts["hub-name"], opts["hub-version"]]
    template = index_templator(parts, opts)
    return template


def lookup_taxon_by_taxid(es, taxon_id, taxonomy_template):
    """Lookup taxon in taxonomy index by taxon_id."""
    query = EsQueryBuilder()
    query.es_match("taxon_id", taxon_id)
    with tolog.DisableLogger():
        res = es.search(index=taxonomy_template["index_name"], body=query.write())
    if res["hits"]["total"]["value"] == 1:
        return res["hits"]["hits"][0]["_source"]
    return None


def lookup_taxa_by_taxon_id(es, values, template, *, return_type="list"):
    """Retrieve existing taxa from index."""
    taxa = []
    if return_type == "dict":
        taxa = {}
    res = document_by_id(
        es, ["taxon_id-%s" % value for value in values], template["index_name"]
    )
    if res is not None:
        if return_type == "list":
            taxa = [key.replace("taxon_id-", "") for key in res.keys()]
        else:
            for key, source in res.items():
                taxon_id = key.replace("taxon_id-", "")
                taxa.update({taxon_id: {"_id": key, "_source": source}})
    # taxon_res = query_keyword_value_template(
    #     es,
    #     "attributes_names_by_keyword_value",
    #     "taxon_id",
    #     values,
    #     index=template["index_name"],
    # )
    # if taxon_res is not None:
    #     for response in taxon_res["responses"]:
    #         if "hits" in response and response["hits"]["total"]["value"] == 1:
    #             if return_type == "list":
    #                 taxa.append(response["hits"]["hits"][0])
    #             else:
    #                 taxa[response["hits"]["hits"][0]["_source"]["taxon_id"]] = response[
    #                     "hits"
    #                 ]["hits"][0]
    #         elif return_type == "list":
    #             taxa.append(None)
    return taxa


def lookup_missing_taxon_ids(
    es,
    without_ids,
    opts,
    *,
    with_ids=None,
    blanks=set(["NA", "None"]),
    spellings=None,
    taxon_table=None,
):
    """Lookup taxon ID based on available taxonomic information."""
    if with_ids is None:
        with_ids = {}
    if spellings is None:
        spellings = {"spellcheck": {}, "synonym": {}}
    # TODO: set this list from types file
    ranks = [
        "subspecies",
        "species",
        "genus",
        "family",
        "order",
        "class",
        "subphylum",
        "phylum",
    ]
    found_keys = []
    pbar = tqdm(total=len(without_ids.keys()))
    for key, arr in without_ids.items():
        pbar.update(1)
        for obj in arr:
            if "taxonomy" not in obj:
                continue
            for index, rank in enumerate(ranks):
                if rank not in obj["taxonomy"] or obj["taxonomy"][rank] in blanks:
                    continue
                taxa, name_class = lookup_taxon(
                    es,
                    obj["taxonomy"][rank],
                    opts,
                    rank=rank,
                    return_type="taxon",
                    spellings=spellings,
                    taxon_table=taxon_table,
                )
                if index == 1 and not taxa:
                    break
                if len(taxa) == 1:
                    obj.update({"input_name": obj["taxonomy"][rank]})
                    taxon = taxa[0]["_source"]
                    if obj["taxonomy"][rank] != taxon["scientific_name"]:
                        spellings["synonym"].update(
                            {
                                obj["taxonomy"][rank]: {
                                    "matches": [taxon["scientific_name"]],
                                    "taxon_id": taxon["taxon_id"],
                                    "rank": rank,
                                }
                            }
                        )
                    if taxon["taxon_id"] in with_ids:
                        with_ids[taxon["taxon_id"]].append(obj)
                    else:
                        obj["attributes"] = [obj["attributes"]]
                        with_ids[taxon["taxon_id"]] = [obj]
                        LOGGER.debug(
                            "Matched %s with taxon_id %s",
                            obj["taxonomy"][rank],
                            taxon["taxon_id"],
                        )
                    found_keys.append(key)
                break
                # TODO: Allow lookup for higher taxa
                # if not taxon_ids:
                #     continue
                for anc_rank in ranks[(index + 1) :]:
                    if (
                        anc_rank not in obj["taxonomy"]
                        or obj["taxonomy"][anc_rank] in blanks
                    ):
                        continue
                    taxon_ids = lookup_taxon_within_lineage(
                        es,
                        obj["taxonomy"][rank],
                        obj["taxonomy"][anc_rank],
                        opts,
                        rank=rank,
                        anc_rank=anc_rank,
                        return_type="taxon_id",
                        name_class=name_class,
                    )
                    if taxon_ids:
                        if len(taxon_ids) == 1:
                            if taxon_ids[0] in with_ids:
                                with_ids[taxon_ids[0]].append(obj)
                            else:
                                obj["attributes"] = [obj["attributes"]]
                                with_ids[taxon_ids[0]] = [obj]
                                LOGGER.debug(
                                    "Matched %s with taxon_id %s",
                                    obj["taxonomy"][rank],
                                    taxon_ids[0],
                                )
                            found_keys.append(key)
                        else:
                            LOGGER.warn(
                                "Taxon name %s is ambiguous within %s",
                                obj["taxonomy"][rank],
                                obj["taxonomy"][anc_rank],
                            )
                        break
                break
    pbar.close()
    found_ids = {taxon_id: True for taxon_id in with_ids.keys()}
    for key in found_keys:
        without_ids.pop(key, None)
        found_ids[key] = True
    return with_ids, without_ids, found_ids


def stream_taxon_names(es, *, index, root=None, size=1000):
    """Get entries by depth of root taxon."""
    if root is not None:
        body = {
            "id": "taxon_names_by_root",
            "params": {"root": root},
        }
        return stream_template_search_results(es, index=index, body=body, size=size)
    body = {
        "id": "taxon_names",
        "params": {},
    }
    return stream_template_search_results(es, index=index, body=body)


def chunker(seq, size):
    """Loop through array in chunks."""
    return (seq[pos : pos + size] for pos in range(0, len(seq), size))


def translate_xrefs(es, *, index, xrefs, source):
    """Translate a list of xrefs into taxon_ids."""
    id_map = {}
    for refs in chunker(xrefs, 20):
        responses = query_keyword_value_template(
            es,
            "taxon_by_specific_name",
            source,
            refs,
            index,
            opts={"keyword": "source", "value": "name"},
        )
        for idx, res in enumerate(responses["responses"]):
            if "hits" in res and "hits" in res["hits"]:
                hits = res["hits"]["hits"]
                if len(hits) == 1:
                    id_map[refs[idx]] = hits[0]["_source"]["taxon_id"]
    return id_map


def load_taxon_table(es, opts, taxonomy_name, taxon_table):
    """Load all taxa into memory for taxon name lookup and spellcheck."""
    LOGGER.info("Loading taxa into memory for taxon name lookup")
    taxon_template = index_template(taxonomy_name, opts)
    root = None
    if "taxon-lookup-root" in opts:
        root = opts["taxon-lookup-root"]
    for node in tqdm(
        stream_taxon_names(es, index=taxon_template["index_name"], root=root)
    ):
        lineage = {}
        node_names = set()
        try:
            if "attributes" in node["_source"]:
                attributes = node["_source"]["attributes"]
            else:
                attributes = []
            for anc in node["_source"]["lineage"]:
                lineage.update({anc["taxon_rank"]: anc["scientific_name"]})
            taxon = {
                "taxon_id": node["_source"]["taxon_id"],
                "taxon_rank": node["_source"]["taxon_rank"],
                "scientific_name": node["_source"]["scientific_name"],
                "lineage": lineage,
                "attributes": attributes,
            }
            taxon_table["scientific"][node["_source"]["scientific_name"]].append(taxon)
            taxon_table["any"][node["_source"]["scientific_name"]].append(taxon)
            node_names.add(node["_source"]["scientific_name"])
            for obj in node["_source"]["taxon_names"]:
                if obj["name"] not in node_names:
                    node_names.add(obj["name"])
                    taxon_table["any"][obj["name"]].append(taxon)
        except KeyError:
            pass


def fix_missing_ids(
    es,
    opts,
    without_ids,
    *,
    types,
    taxon_template,
    failed_rows,
    imported_rows,
    with_ids=None,
    blanks=set(["NA", "None"]),
    header=None,
    spellings=None,
    taxon_table=None,
):
    """Find or create taxon IDs for rows without."""
    if with_ids is None:
        with_ids = {}
    if spellings is None:
        spellings = {"spellcheck": {}, "synonym": {}}
    found_ids = {}
    if without_ids:
        # TODO: support multiple taxonomies
        LOGGER.info("Looking up %d missing taxon IDs", len(without_ids.keys()))
        with_ids, without_ids, found_ids = lookup_missing_taxon_ids(
            es,
            without_ids,
            opts,
            with_ids=with_ids,
            blanks=blanks,
            spellings=spellings,
            taxon_table=taxon_table,
        )
        # create new taxon IDs
        if "taxonomy" in types and "alt_taxon_id" in types["taxonomy"]:
            LOGGER.info(
                "Using alt_taxon_id to fill in %d missing taxon IDs",
                len(without_ids.keys()),
            )
            created_ids = create_taxa(
                es,
                opts,
                data=without_ids,
                blanks=blanks,
                taxon_template=taxon_template,
                spellings=spellings,
            )
            for created_id in created_ids:
                if created_id in without_ids:
                    with_ids[created_id] = without_ids[created_id]
                    found_ids[created_id] = True
                    del without_ids[created_id]
    if failed_rows:
        for key, value in found_ids.items():
            if key in failed_rows:
                imported_rows += failed_rows[key]
                del failed_rows[key]
        if failed_rows:
            LOGGER.info(
                "Unable to associate %d records with taxon IDs", len(failed_rows)
            )
            write_imported_rows(
                failed_rows, opts, types=types, header=header, label="exceptions"
            )
    return with_ids, without_ids


def add_taxonomy_info_to_meta(meta, source):
    """Add taxonomy information to a metadata dict."""
    keys = {"lineage", "parent", "scientific_name", "taxon_names", "taxon_rank"}
    for key in keys:
        if key in source:
            meta[key] = source[key]


def stream_taxa(taxa):
    """Stream dict of taxa for indexing."""
    for taxon_id, value in taxa.items():
        yield "taxon_id-%s" % taxon_id, value


def get_taxa_to_create(
    es,
    opts,
    *,
    taxonomy_name="ncbi",
    taxon_ids=None,
    asm_by_taxon_id=None,
):
    """Create a dict of taxa to create."""
    taxa_to_create = {}
    if not taxon_ids:
        return {}
    if asm_by_taxon_id is None:
        asm_by_taxon_id = {}
    taxonomy_template = taxonomy_index_template(taxonomy_name, opts)
    taxonomy_res = query_value_template(
        es,
        "taxonomy_node_by_taxon_id",
        taxon_ids,
        taxonomy_template["index_name"],
    )
    if taxonomy_res is None:
        LOGGER.error(
            "Could not connect to taxonomy index '%s'",
            taxonomy_template["index_name"],
        )
        sys.exit(1)
    ancestors = set()
    for taxonomy_result in taxonomy_res["responses"]:
        if taxonomy_result["hits"]["total"]["value"] == 1:
            source = taxonomy_result["hits"]["hits"][0]["_source"]
            taxa_to_create[source["taxon_id"]] = source
            for ancestor in source["lineage"]:
                ancestors.add(ancestor["taxon_id"])
            if source["taxon_id"] in asm_by_taxon_id:
                for asm in asm_by_taxon_id[source["taxon_id"]]:
                    add_taxonomy_info_to_meta(asm, source)
    taxonomy_res = query_value_template(
        es,
        "taxonomy_node_by_taxon_id",
        list(ancestors),
        taxonomy_template["index_name"],
    )
    if taxonomy_res and "responses" in taxonomy_res:
        for taxonomy_result in taxonomy_res["responses"]:
            if taxonomy_result["hits"]["total"]["value"] == 1:
                source = taxonomy_result["hits"]["hits"][0]["_source"]
                taxa_to_create[source["taxon_id"]] = source
    return taxa_to_create


def find_or_create_taxa(es, opts, *, taxon_ids, taxon_template, asm_by_taxon_id=None):
    """Find taxa in taxon index or create new taxon entries from taxonomy."""
    taxa = lookup_taxa_by_taxon_id(
        es, list(taxon_ids), taxon_template, return_type="dict"
    )
    missing_taxa = [taxon_id for taxon_id in taxon_ids if taxon_id not in taxa]
    to_create = get_taxa_to_create(
        es,
        opts,
        taxonomy_name="ncbi",
        taxon_ids=missing_taxa,
        asm_by_taxon_id=asm_by_taxon_id,
    )
    index_stream(
        es,
        taxon_template["index_name"],
        stream_taxa(to_create),
        dry_run=opts.get("dry-run", False),
    )
    taxa.update(
        {
            taxon_id: {"_id": "taxon_id-%s" % taxon_id, "_source": obj}
            for taxon_id, obj in to_create.items()
        }
    )
    return taxa


def list_ancestors(taxa):
    """Convert a dict of taxa to a list of ancestral taxon_ids."""
    ancestors = set()
    for doc in taxa.values():
        if "_source" in doc:
            lineage = doc["_source"].get("lineage", [])
        else:
            lineage = doc.get("lineage", [])
        for entry in lineage:
            ancestors.add(entry["taxon_id"])
    return list(ancestors)


def add_names_to_list(existing, new, *, blanks=set({"NA", "None"})):
    """Add names to a list if they do not already exist."""
    names = defaultdict(dict)
    for entry in existing:
        names[entry["class"]][entry["name"]] = True
    for entry in new:
        entry["class"] = entry["class"].lower()  # .replace("_", " ")
        if entry["name"] not in blanks and entry["name"] not in names[entry["class"]]:
            existing.append(entry)
            names[entry["class"]][entry["name"]] = True


def add_names_and_attributes_to_taxa(
    es, data, opts, *, template, blanks=set(["NA", "None"])
):
    """Add names and attributes to taxa."""
    for values in chunks(list(data.keys()), 500):
        # taxa = lookup_taxa_by_taxon_id(es, values, template, return_type="list")
        all_taxa = find_or_create_taxa(
            es,
            opts,
            taxon_ids=values,
            taxon_template=template,
        )
        taxa = []
        for taxon_id in values:
            if taxon_id in all_taxa:
                taxa.append(all_taxa[taxon_id])
        for doc in taxa:
            if doc is not None:
                taxon_data = data[doc["_source"]["taxon_id"]]
                taxon_names = []
                attributes = []
                for entry in taxon_data:
                    if "attributes" in entry:
                        attributes = attributes + entry["attributes"]
                    if "taxon_names" in entry:
                        taxon_names += entry["taxon_names"]
                if "taxon_names" not in doc["_source"]:
                    doc["_source"]["taxon_names"] = []
                add_names_to_list(
                    doc["_source"]["taxon_names"], taxon_names, blanks=blanks
                )
                if (
                    "attributes" not in doc["_source"]
                    or not doc["_source"]["attributes"]
                ):
                    doc["_source"]["attributes"] = []
                add_attribute_values(doc["_source"]["attributes"], attributes)
                yield doc["_id"], doc["_source"]


def lookup_taxon_within_lineage(
    es,
    name,
    lineage,
    opts,
    *,
    rank=None,
    anc_rank=None,
    return_type="taxon",
    name_class="scientific",
):
    """Lookup taxon ID in a specified lineage."""
    template = index_template(opts["taxonomy-source"].lower(), opts)
    body = {
        "id": "taxon_by_lineage",
        "params": {
            "taxon": name,
            "rank": rank,
            "lineage": lineage,
            "anc_rank": anc_rank,
        },
    }
    if name_class == "any":
        body.update({"id": "taxon_by_any_name_by_lineage"})
    with tolog.DisableLogger():
        res = es.search_template(
            body=body, index=template["index_name"], rest_total_hits_as_int=True
        )
    if "hits" in res and res["hits"]["total"] > 0:
        if return_type == "taxon_id":
            return [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
        return [hit for hit in res["hits"]["hits"]]
    index = template["index_name"].replace("taxon", "taxonomy")
    with tolog.DisableLogger():
        res = es.search_template(body=body, index=index, rest_total_hits_as_int=True)
    if "hits" in res and res["hits"]["total"] > 0:
        if return_type == "taxon_id":
            return [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
        return [hit for hit in res["hits"]["hits"]]
    return []


def spellcheck_taxon(es, name, index, rank, taxonomy_index_template, opts, return_type):
    """Look up taxon name with fuzzy matching."""
    taxon_suggest = {
        "id": "taxon_suggest",
        "params": {"searchTerm": name},
    }
    matches = None
    with tolog.DisableLogger():
        suggestions = es.search_template(
            body=taxon_suggest, index=index, rest_total_hits_as_int=True
        )
        try:
            options = suggestions["suggest"]["simple_phrase"][0]["options"]
            matches = [
                option["text"]
                for option in options
                if option.get("collate_match", False)
            ]
        except KeyError:
            return None, rank, None
        except ValueError:
            return None, rank, None
    taxon_id = None
    if matches:
        taxon_matches = {}
        scientific_name = None
        for match in matches:
            body = {
                "id": "taxon_by_any_name",
                "params": {"taxon": match, "rank": rank},
            }
            taxa = taxon_lookup(
                es, body, index, taxonomy_index_template, opts, return_type="taxon"
            )
            if len(taxa) > 1:
                return None, rank, matches
            for taxon in taxa:
                source = taxon["_source"]
                taxon_id = source["taxon_id"]
                taxon_matches[taxon_id] = source["scientific_name"]
                scientific_name = source["scientific_name"]
        if len(taxon_matches.keys()) == 1:
            return taxon_id, rank, [scientific_name]
    return None, rank, matches


def taxon_lookup(es, body, index, taxonomy_index_template, opts, return_type):
    """Query elasticsearch for a taxon."""
    taxa = []
    with tolog.DisableLogger():
        res = es.search_template(body=body, index=index, rest_total_hits_as_int=True)
    if "hits" in res and res["hits"]["total"] > 0:
        if return_type == "taxon_id":
            taxa = [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
        else:
            taxa = [hit for hit in res["hits"]["hits"]]
    else:
        template = taxonomy_index_template(opts["taxonomy-source"].lower(), opts)
        index = template["index_name"]
        with tolog.DisableLogger():
            res = es.search_template(
                body=body, index=index, rest_total_hits_as_int=True
            )
        if "hits" in res and res["hits"]["total"] > 0:
            if return_type == "taxon_id":
                taxa = [hit["_source"]["taxon_id"] for hit in res["hits"]["hits"]]
            else:
                taxa = [hit for hit in res["hits"]["hits"]]
    return taxa


def lookup_taxon_in_index(
    es,
    name,
    opts,
    *,
    rank,
    name_class,
    return_type,
    spellings,
):
    """Lookup taxon in Elasticsearch index."""
    template = index_template(opts["taxonomy-source"].lower(), opts)
    index = template["index_name"]
    body = {
        "id": "taxon_by_name",
        "params": {"taxon": name, "rank": rank},
    }
    if name_class in {"any", "spellcheck"}:
        body.update({"id": "taxon_by_any_name"})
    if name_class == "spellcheck":
        taxon_id, rank, matches = spellcheck_taxon(
            es, name, index, rank, taxonomy_index_template, opts, return_type
        )
        if matches:
            spellings["spellcheck"].update(
                {name: {"matches": matches, "taxon_id": taxon_id, "rank": rank}}
            )
        return [], name_class
    taxa = taxon_lookup(es, body, index, taxonomy_index_template, opts, return_type)
    return taxa


def lookup_taxon_in_memory(
    name, opts, *, rank, name_class, return_type, spellings, taxon_table
):
    """Lookup taxon in memory."""
    taxa = []
    if name_class in taxon_table:
        if name in taxon_table[name_class]:
            for obj in taxon_table[name_class][name]:
                if return_type == "taxon_id":
                    taxa.append(obj["taxon_id"])
                else:
                    taxa.append({"_source": {**obj}})
    return taxa


def lookup_taxon(
    es,
    name,
    opts,
    *,
    rank=None,
    name_class="scientific",
    return_type="taxon_id",
    spellings=None,
    taxon_table=None,
):
    """Lookup taxon ID."""
    if spellings is None:
        spellings = {"spellcheck": {}, "synonym": {}}
    if taxon_table is None or name_class == "spellcheck":
        taxa = lookup_taxon_in_index(
            es,
            name,
            opts,
            rank=rank,
            name_class=name_class,
            return_type=return_type,
            spellings=spellings,
        )
    else:
        taxa = lookup_taxon_in_memory(
            name,
            opts,
            rank=rank,
            name_class=name_class,
            return_type=return_type,
            spellings=spellings,
            taxon_table=taxon_table,
        )
    if (
        not taxa
        and opts["taxon-lookup"] == "any"
        and name_class not in {"any", "spellcheck"}
    ):
        taxa, name_class = lookup_taxon(
            es,
            name,
            opts,
            rank=rank,
            name_class="any",
            return_type=return_type,
            spellings=spellings,
            taxon_table=taxon_table,
        )
    if (
        not taxa
        and "taxon-spellcheck" in opts
        and opts["taxon-spellcheck"]
        and name_class != "spellcheck"
    ):
        taxa, name_class = lookup_taxon(
            es,
            name,
            opts,
            rank=rank,
            name_class="spellcheck",
            return_type=return_type,
            spellings=spellings,
            taxon_table=taxon_table,
        )
    return taxa, name_class


def generate_ancestral_taxon_id(name, rank, *, alt_taxon_id=None, taxon_ids=None):
    """Generate an ancestral taxon ID."""
    if taxon_ids is None:
        taxon_ids = set({})
    increment = 0
    while True:
        # TODO: make robust to imports from separate files
        anc_taxon_id = "anc_%s" % name
        if increment:
            anc_taxon_id += "_%d" % increment
        if anc_taxon_id not in taxon_ids:
            taxon_ids.add(anc_taxon_id)
            return anc_taxon_id
        increment += 1


def create_descendant_taxon(taxon_id, rank, name, closest_taxon):
    """Set taxon and lineage information for a descendant taxon."""
    desc_taxon = {
        "_id": "taxon_id-%s" % taxon_id,
        "_source": {
            "taxon_id": taxon_id,
            "taxon_rank": rank,
            "scientific_name": name,
            "parent": closest_taxon["_source"]["taxon_id"],
            "taxon_names": [{"class": "scientific name", "name": name}],
        },
    }
    lineage = [
        {
            "node_depth": 1,
            "taxon_id": closest_taxon["_source"]["taxon_id"],
            "taxon_rank": closest_taxon["_source"]["taxon_rank"],
            "scientific_name": closest_taxon["_source"]["scientific_name"],
        }
    ]
    if "lineage" in closest_taxon["_source"]:
        for ancestor in closest_taxon["_source"]["lineage"]:
            lineage.append(
                {
                    "taxon_id": ancestor["taxon_id"],
                    "taxon_rank": ancestor["taxon_rank"],
                    "scientific_name": ancestor["scientific_name"],
                    "node_depth": ancestor["node_depth"] + 1,
                }
            )
    desc_taxon["_source"]["lineage"] = lineage
    return desc_taxon


def add_new_taxon(alt_taxon_id, new_taxa, obj, closest_taxon, *, blanks={"NA", "None"}):
    """Add a new taxon with alt_taxon_id to list of taxa."""
    # TODO: Allow creation of new higher taxa
    ranks = [
        "subspecies",
        "species",
        "genus",
        "family",
        "order",
        "class",
        "subphylum",
        "phylum",
    ]
    if alt_taxon_id not in new_taxa:
        alt_rank = None
        for rank in ranks:
            if (
                rank in obj["taxonomy"]
                and obj["taxonomy"][rank]
                and obj["taxonomy"][rank] not in blanks
            ):
                alt_rank = rank
                break
        if alt_rank is not None:
            new_taxon = create_descendant_taxon(
                alt_taxon_id, alt_rank, obj["taxonomy"][alt_rank], closest_taxon
            )
            new_taxa.update({new_taxon["_source"]["taxon_id"]: new_taxon["_source"]})
    return new_taxon


def set_ranks(taxonomy):
    """Set ranks for species/subspecies creation."""
    default_ranks = [
        "genus",
        "family",
        "order",
        "class",
        "subphylum",
        "phylum",
    ]
    taxon_rank = None
    if "subspecies" in taxonomy:
        ranks = ["species"] + default_ranks
        taxon_rank = "subspecies"
    else:
        ranks = default_ranks
        for rank in ["species"] + default_ranks:
            if rank in taxonomy:
                taxon_rank = rank
                break
    return ranks, taxon_rank


def create_new_taxon(
    alt_taxon_id,
    closest_taxon,
    closest_rank,
    lineage,
    new_taxa,
    taxon_ids,
    obj,
    matches,
    taxa,
    ancestors,
):
    """Create a new taxon with new ancestral taxa as required."""
    if closest_taxon is not None:
        for intermediate in reversed(lineage):
            taxon_id = generate_ancestral_taxon_id(
                intermediate["name"],
                intermediate["rank"],
                alt_taxon_id=alt_taxon_id,
                taxon_ids=taxon_ids,
            )
            new_taxon = create_descendant_taxon(
                taxon_id, intermediate["rank"], intermediate["name"], closest_taxon
            )
            new_taxa.update({new_taxon["_source"]["taxon_id"]: new_taxon["_source"]})
            matches[intermediate["name"]][obj["taxonomy"][closest_rank]] = taxa
            closest_rank = intermediate["rank"]
            closest_taxon = new_taxon
        ancestors[alt_taxon_id] = closest_taxon
        added_taxon = add_new_taxon(alt_taxon_id, new_taxa, obj, closest_taxon)
        matches[added_taxon["_source"]["scientific_name"]][
            closest_taxon["_source"]["scientific_name"]
        ] = [added_taxon]


def find_ancestor(
    ranks,
    index,
    obj,
    taxon,
    matches,
    ancestors,
    alt_taxon_id,
    es,
    opts,
    rank,
    lookup_name_class,
    intermediates,
    blanks,
):
    """Loop through ancestral ranks to link new taxon to an ancestor."""
    for anc_rank in ranks[(index + 1) :]:
        if anc_rank not in obj["taxonomy"] or obj["taxonomy"][anc_rank] in blanks:
            # row has no name at this rank
            continue
        if taxon in matches and obj["taxonomy"][anc_rank] in matches[taxon]:
            # this taxon has been seen before
            taxa = matches[taxon][obj["taxonomy"][anc_rank]]
            ancestors.update({alt_taxon_id: taxa[0]})
            break
        else:
            # find existing ancestral taxa within a lineage
            # TODO: make an in memory version of this lookup
            taxa = lookup_taxon_within_lineage(
                es,
                taxon,
                obj["taxonomy"][anc_rank],
                opts,
                rank=rank,
                anc_rank=anc_rank,
                return_type="taxon",
                name_class=lookup_name_class,
            )
        if taxa:
            if len(taxa) == 1:
                #  unambiguous match to a single existing taxon
                ancestors.update({alt_taxon_id: taxa[0]})
                matches[taxon][obj["taxonomy"][anc_rank]] = taxa
                break
        elif intermediates == 0:
            taxa, name_class = lookup_taxon(
                es,
                obj["taxonomy"][anc_rank],
                opts,
                rank=anc_rank,
                return_type="taxon",
                name_class="any",
            )
            if taxa and len(taxa) == 1:
                taxa = lookup_taxon_within_lineage(
                    es,
                    taxon,
                    taxa[0]["_source"]["scientific_name"],
                    opts,
                    rank=rank,
                    anc_rank=anc_rank,
                    return_type="taxon",
                    name_class=lookup_name_class,
                )
                if taxa and len(taxa) == 1:
                    ancestors.update({alt_taxon_id: taxa[0]})
                    matches[taxon][obj["taxonomy"][anc_rank]] = taxa
                    break
        intermediates += 1
    return anc_rank, taxa


def create_taxa(
    es, opts, *, taxon_template, data=None, blanks=set(["NA", "None"]), spellings=None
):
    """Create new taxa using alternate taxon IDs."""
    if spellings is None:
        spellings = {"spellcheck": {}, "synonym": {}}
    ancestors = {}
    matches = defaultdict(dict)
    pbar = tqdm(total=len(data.keys()))
    taxon_ids = set({})
    new_taxa = {}
    for rows in data.values():
        obj = rows[0]
        pbar.update(1)
        if (
            "taxonomy" not in obj
            or "alt_taxon_id" not in obj["taxonomy"]
            or obj["taxonomy"]["alt_taxon_id"] in blanks
        ):
            # row has no alt_taxon_id
            continue
        alt_taxon_id = obj["taxonomy"]["alt_taxon_id"]
        lineage = []
        closest_rank = None
        closest_taxon = None
        # fetch ancestral ranks and current taxon rank
        ranks, taxon_rank = set_ranks(obj["taxonomy"])
        if taxon_rank in obj["taxonomy"] and obj["taxonomy"][taxon_rank] in spellings:
            # taxon name may be mis-spelled
            continue
        max_index = len(ranks) - 1
        lookup_name_class = "any" if opts["taxon-lookup"] == "any" else "scientific"
        # loop through lineage to find existing ancestral taxa
        for index, rank in enumerate(ranks[: (max_index - 1)]):
            if rank not in obj["taxonomy"] or obj["taxonomy"][rank] in blanks:
                # row has no name at this rank
                continue
            if obj["taxonomy"][rank] in spellings:
                # ancestral taxon name is missing or may be mis-spelled
                break
            intermediates = 0
            taxon = obj["taxonomy"][rank]
            # loop through higher ranks to disambiguate name clashes
            anc_rank, taxa = find_ancestor(
                ranks,
                index,
                obj,
                taxon,
                matches,
                ancestors,
                alt_taxon_id,
                es,
                opts,
                rank,
                lookup_name_class,
                intermediates,
                blanks,
            )
            if alt_taxon_id in ancestors:
                closest_rank = rank
                if taxon in matches and obj["taxonomy"][anc_rank] in matches[taxon]:
                    closest_taxon = matches[taxon][obj["taxonomy"][anc_rank]][0]
                else:
                    closest_taxon = matches[obj["taxonomy"][anc_rank]]["all"][0]
                break
            lineage.append({"rank": rank, "name": taxon})
        # create a new taxon if a closest ancestral taxon could be found
        create_new_taxon(
            alt_taxon_id,
            closest_taxon,
            closest_rank,
            lineage,
            new_taxa,
            taxon_ids,
            obj,
            matches,
            taxa,
            ancestors,
        )
    pbar.close()
    # add new taxa to the index
    index_stream(
        es,
        taxon_template["index_name"],
        stream_taxa(new_taxa),
        dry_run=opts.get("dry-run", False),
    )
    # return a list of alt_taxon_ids for the created taxa
    return new_taxa.keys()



#!/usr/bin/env python3

"""
Index a file, directory or repository.

Usage:
    genomehubs index [--hub-name STRING] [--hub-path PATH] [--hub-version PATH]
                     [--config-file PATH...] [--config-save PATH]
                     [--es-host URL...] [--assembly-dir PATH]
                     [--assembly-repo URL] [--assembly-exception PATH]
                     [--taxon-dir PATH] [--taxon-repo URL] [--taxon-exception PATH]
                     [--taxon-lookup STRING] [--taxon-lookup-root STRING]
                     [--taxon-lookup-in-memory] [--taxon-id-as-xref STRING]
                     [--taxon-spellcheck] [--taxonomy-source STRING]
                     [--file PATH...] [file-dir PATH...]
                     [--remote-file URL...] [--remote-file-dir URL...]
                     [--taxon-id STRING] [--assembly-id STRING] [--analysis-id STRING]
                     [--file-title STRING] [--file-description STRING] [--file-metadata PATH]
                     [--dry-run]
                     [-h|--help] [-v|--version]

Options:
    --hub-name STRING          GenomeHubs instance name.
    --hub-path PATH            GenomeHubs instance root directory path.
    --hub-version STR          GenomeHubs instance version string.
    --config-file PATH         Path to YAML file containing configuration options.
    --config-save PATH         Path to write configuration options to YAML file.
    --es-host URL              ElasticSearch hostname/URL and port.
    --assembly-dir PATH        Path to directory containing assembly-level data.
    --assembly-repo URL        Remote git repository containing assembly-level data.
                               Optionally include `~branch-name` suffix.
    --assembly-exception PATH  Path to directory to write assembly data that failed to import.
    --taxon-lookup-root STRING Root taxon Id for in-memory lookup.
    --taxon-lookup STRING      Taxon name class to lookup (scientific|any). [Default: scientific]
    --taxon-lookup-in-memory   Flag to use in-memory taxon name lookup.
    --taxon-id-as-xref STRING  Set source DB name to treat taxon_id in file as xref.
    --taxon-spellcheck         Flag to use fuzzy matching to match taxon names.
    --taxon-dir PATH           Path to directory containing taxon-level data.
    --taxon-repo URL           Remote git repository containing taxon-level data.
                               Optionally include `~branch-name` suffix.
    --taxon-exception PATH     Path to directory to write taxon data that failed to import.
    --taxonomy-source STRING   Name of taxonomy to use (ncbi or ott).
    --file PATH                Path to file for generic file import.
    --file-dir PATH            Path to directory containing generic files to import.
    --remote-file URL          Location of remote file for generic file import.
    --remote-file-dir URL      Location of remote directory containing generic files to import.
    --taxon-id STRING          Taxon ID to index files against.
    --assembly-id STRING       Assembly ID to index files against.
    --analysis-id STRING       Analysis ID to index files against.
    --file-title STRING        Default title for indexed files.
    --file-description STRING  Default description for all indexed files.
    --file-metadata PATH       CSV, TSV, YAML or JSON file metadata with one entry per file to be indexed.
    --dry-run                  Flag to run without loading data into the elasticsearch index.
    -h, --help                 Show this
    -v, --version              Show version number

Examples:
    # 1. Index all files in a remote repository
    ./genomehubs index --taxon-repo https://github.com/genomehubs/goat-data
"""

import csv
import sys
from collections import defaultdict
from pathlib import Path

from docopt import docopt
from tolkein import tolog
from tqdm import tqdm

from ..lib import assembly
from ..lib import es_functions
from ..lib import hub
from ..lib import taxon
from .assembly import add_identifiers_and_attributes_to_assemblies
from .attributes import index_types
from .config import config
from .es_functions import index_stream
from .files import index_files
from .files import index_metadata
from .hub import process_row
from .hub import set_column_indices
from .hub import strip_comments
from .hub import validate_types_file
from .hub import write_imported_rows
from .hub import write_imported_taxa
from .hub import write_spellchecked_taxa
from .taxon import add_names_and_attributes_to_taxa
from .taxon import fix_missing_ids
from .taxon import load_taxon_table
from .taxon import translate_xrefs
from .version import __version__

LOGGER = tolog.logger(__name__)


def not_blank(key, obj, blanks):
    """Test value is not blank."""
    return key in obj and obj[key] and obj[key] not in blanks


def summarise_imported_taxa(docs, imported_taxa):
    """Summarise taxon imformation from a stram of taxon docs."""
    for entry_id, entry in docs:
        imported_taxa[entry["scientific_name"]].append(
            {
                "taxon_id": entry["taxon_id"],
                "rank": entry["taxon_rank"],
            }
        )
        yield entry_id, entry


def group_rows(
    taxon_id,
    rows,
    with_ids,
    without_ids,
    taxon_asm_data,
    imported_rows,
    types,
    failed_rows,
    blanks,
):
    """Group processed rows by available taxon info for import."""
    ranks = ["subspecies", "species", "family", "order", "class"]
    for processed_data, taxon_data, row in rows:
        if taxon_id != "other":
            with_ids[taxon_id].append(processed_data)
            taxon_asm_data[taxon_id].append(taxon_data)
            imported_rows.append(row)

        else:
            if "taxonomy" in types and not_blank(
                "alt_taxon_id", processed_data["taxonomy"], blanks
            ):
                without_ids[processed_data["taxonomy"]["alt_taxon_id"]].append(
                    processed_data
                )
                taxon_asm_data[processed_data["taxonomy"]["alt_taxon_id"]].append(
                    taxon_data
                )
                failed_rows[processed_data["taxonomy"]["alt_taxon_id"]].append(row)
            else:
                row_rank = None
                for rank in ranks:
                    if not_blank(rank, processed_data["taxonomy"], blanks):
                        row_rank = rank
                        without_ids[processed_data["taxonomy"][rank]].append(
                            processed_data
                        )
                        taxon_asm_data[processed_data["taxonomy"][rank]].append(
                            taxon_data
                        )
                        failed_rows[processed_data["taxonomy"][rank]].append(row)
                        break
                if row_rank is None:
                    failed_rows["None"].append(row)


def index_file(es, types, names, data, opts, *, taxon_table=None):
    """Index a file."""
    delimiters = {"csv": ",", "tsv": "\t"}
    rows = csv.reader(
        strip_comments(data, types),
        delimiter=delimiters[types["file"]["format"]],
        quotechar='"',
    )
    if "header" in types["file"] and types["file"]["header"]:
        header = next(rows)
        set_column_indices(types, header)
    else:
        header = None
    with_ids = defaultdict(list)
    taxon_asm_data = defaultdict(list)
    without_ids = defaultdict(list)
    failed_rows = defaultdict(list)
    imported_rows = []
    blanks = set(["", "NA", "N/A", "None"])
    taxon_types = {}
    taxonomy_name = opts["taxonomy-source"].lower()
    taxon_template = taxon.index_template(taxonomy_name, opts)
    LOGGER.info("Processing rows")
    processed_rows = defaultdict(list)
    for row in tqdm(rows):
        try:
            processed_data, taxon_data, new_taxon_types = process_row(types, names, row)
        except Exception as err:
            print(err)
            failed_rows["None"].append(row)
            continue
        taxon_types.update(new_taxon_types)
        if not_blank("_taxon_id", processed_data["taxonomy"], blanks):
            # if opts["taxon-id-as-xref"]:
            with_ids[processed_data["taxonomy"]["_taxon_id"]].append(processed_data)
            taxon_asm_data[processed_data["taxonomy"]["_taxon_id"]].append(taxon_data)
            imported_rows.append(row)
        else:
            tmp_taxon_id = "other"
            if not_blank("taxon_id", processed_data["taxonomy"], blanks):
                tmp_taxon_id = processed_data["taxonomy"]["taxon_id"]
            processed_rows[tmp_taxon_id].append((processed_data, taxon_data, row))
    # check for taxon-id-as-xref here and do lookup if required
    if "taxon-id-as-xref" in opts:
        id_map = translate_xrefs(
            es,
            index=taxon_template["index_name"],
            xrefs=list(processed_rows.keys()),
            source=opts["taxon-id-as-xref"],
        )
        updated_rows = defaultdict(list)
        for xref, taxon_id in id_map.items():
            updated_rows[taxon_id] = processed_rows[xref]
            del processed_rows[xref]
        unmatched = list(processed_rows.keys())
        for xref in unmatched:
            updated_rows["other"] += processed_rows[xref]
            del processed_rows[xref]
    else:
        updated_rows = processed_rows
    for taxon_id, rows in updated_rows.items():
        group_rows(
            taxon_id,
            rows,
            with_ids,
            without_ids,
            taxon_asm_data,
            imported_rows,
            types,
            failed_rows,
            blanks,
        )
    LOGGER.info("Found taxon IDs in %d entries", len(with_ids.keys()))
    spellings = {"spellcheck": {}, "synonym": {}}
    create_ids, without_ids = fix_missing_ids(
        es,
        opts,
        without_ids,
        types=types,
        taxon_template=taxon_template,
        failed_rows=failed_rows,
        imported_rows=imported_rows,
        with_ids=with_ids,
        blanks=blanks,
        header=header,
        spellings=spellings,
        taxon_table=taxon_table,
    )
    write_spellchecked_taxa(spellings, opts, types=types)
    if with_ids or create_ids:
        write_imported_rows(
            imported_rows, opts, types=types, header=header, label="imported"
        )
        LOGGER.info("Indexing %d entries", len(with_ids.keys()))
        if opts["index"] == "taxon":
            docs = add_names_and_attributes_to_taxa(
                es, dict(with_ids), opts, template=taxon_template, blanks=blanks
            )
            imported_taxa = defaultdict(list)
            index_stream(
                es,
                taxon_template["index_name"],
                summarise_imported_taxa(docs, imported_taxa),
                _op_type="update",
                dry_run=opts.get("dry-run", False),
            )
            write_imported_taxa(imported_taxa, opts, types=types)
        elif opts["index"] == "assembly":
            # TODO: keep track of taxon_id not found exceptions
            assembly_template = assembly.index_template(taxonomy_name, opts)
            docs = add_identifiers_and_attributes_to_assemblies(
                es,
                with_ids,
                opts,
                template=assembly_template,
                taxon_template=taxon_template,
                blanks=blanks,
            )
            index_stream(
                es,
                assembly_template["index_name"],
                docs,
                dry_run=opts.get("dry-run", False),
            )
            # index taxon-level attributes
            index_types(
                es,
                "taxon",
                {"attributes": taxon_types},
                opts,
                dry_run=opts.get("dry-run", False),
            )
            taxon_asm_with_ids = {
                taxon_id: taxon_asm_data[taxon_id] for taxon_id in with_ids.keys()
            }
            # print(taxon_asm_with_ids)
            taxon_docs = add_names_and_attributes_to_taxa(
                es,
                taxon_asm_with_ids,
                opts,
                template=taxon_template,
                blanks=blanks,
            )
            index_stream(
                es,
                taxon_template["index_name"],
                taxon_docs,
                _op_type="update",
                dry_run=opts.get("dry-run", False),
            )


def main(args):
    """Index files."""
    options = config("index", **args)

    # Start Elasticsearch
    es = es_functions.launch_es(options["fill"])

    # Post search scripts
    with tolog.DisableLogger():
        hub.post_search_scripts(es)

    taxonomy_name = options["index"]["taxonomy-source"].lower()
    taxon_table = None
    if taxon_table is None and "taxon-lookup-in-memory" in options["index"]:
        taxon_table = {
            "scientific": defaultdict(list),
            "any": defaultdict(list),
        }
        load_taxon_table(es, options["index"], taxonomy_name, taxon_table)
    dry_run = options["index"].get("dry-run", False)
    for index in list(["taxon", "assembly"]):
        data_dir = "%s-dir" % index
        if data_dir in options["index"]:
            dir_path = options["index"][data_dir]
            for types_file in sorted(Path(dir_path).glob("*.names.yaml")):
                types, data, names = validate_types_file(types_file, dir_path)
                LOGGER.info("Indexing %s" % types["file"]["name"])
                index_types(es, index, types, options["index"], dry_run=dry_run)
                index_file(
                    es,
                    types,
                    names,
                    data,
                    {
                        **options["index"],
                        "index": index,
                        "index_types": index_types,
                    },
                    taxon_table=taxon_table,
                )
            for types_file in sorted(Path(dir_path).glob("*.types.yaml")):
                types, data, names = validate_types_file(types_file, dir_path)
                LOGGER.info("Indexing %s" % types["file"]["name"])
                index_types(es, index, types, options["index"], dry_run=dry_run)
                index_file(
                    es,
                    types,
                    names,
                    data,
                    {
                        **options["index"],
                        "index": index,
                        "index_types": index_types,
                    },
                    taxon_table=taxon_table,
                )
    # TODO: #29 Implement alternate backbone taxonomies
    if "file" in options["index"]:
        index_files(es, options["index"]["file"], taxonomy_name, options["index"])
    elif "file-metadata" in options["index"]:
        index_metadata(
            es,
            options["index"]["file-metadata"],
            taxonomy_name,
            options["index"],
            dry_run=dry_run,
        )


def cli():
    """Entry point."""
    if len(sys.argv) == sys.argv.index("index") + 1:
        args = docopt(__doc__, argv=[])
    else:
        args = docopt(__doc__, version=__version__)
    main(args)


if __name__ == "__main__":
    cli()
